var imgFetcher = {
    currentImg: null,
    overImg: null,
    isPopupOpen: false,
    popupText: "",

    overtimeout: null,
    closetimeout: null,

    // those can be overriden by pref
    overdelay: 500,
    closedelay: 300,
    popupPos: 3,

    pref : Components.classes["@mozilla.org/preferences-service;1"]
        .getService(Components.interfaces.nsIPrefService)
        .getBranch("extensions.imgfetcher.options."),
    
    init: function() {
        // get preferences
        try {
            this.overdelay = this.pref.getIntPref("overdelay");
        } catch(e) {}
        try {
            this.closedelay = this.pref.getIntPref("closedelay");
        } catch(e) {}
        try {
            var popupPos = this.pref.getIntPref("popuppos");
            if ([1,2,3,4].some(function(x) { return (popupPos == x) }))
                this.popupPos = popupPos;
        } catch(e) {}

        // observe pref changes
        this.pref.QueryInterface(Components.interfaces.nsIPrefBranch2).addObserver("", this, false);

        // observe zoom modification
        try {
            var cps = Components.classes["@mozilla.org/content-pref/service;1"]
             .getService(Components.interfaces.nsIContentPrefService);
            cps.addObserver("browser.content.full-zoom", this);
        } catch(e) {}
        
        // add dom listeners
        document.addEventListener("mouseout", this.mouseout, true);
        document.addEventListener("mousemove", this.mousemove, true);
        document.addEventListener("mouseover", this.mouseover, true);
        document.getElementById("content").addProgressListener(this, Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);

        // get popup text from stringbundle
        var strBundleSer = Components.classes["@mozilla.org/intl/stringbundle;1"]
            .getService(Components.interfaces.nsIStringBundleService);
        var bundle = strBundleSer.createBundle("chrome://imgfetcher/locale/imgfetcher.properties");
        this.popupText = bundle.GetStringFromName("save");

        // add click handler
        document.getElementById("content").addEventListener("click", this.saveCurrent, true);
    },

    uninit: function() {
        // remove click handler
        document.getElementById("content").removeEventListener("click", this.saveCurrent, true);

        // remove dom listeners
        document.removeEventListener("mouseout", this.mouseout, true);
        document.removeEventListener("mousemove", this.mousemove, true);
        document.removeEventListener("mouseover", this.mouseover, true);
        document.getElementById("content").removeProgressListener(this);

        // unobserve zoom modification
        try {
            var cps = Components.classes["@mozilla.org/content-pref/service;1"]
                .getService(Components.interfaces.nsIContentPrefService);
            cps.removeObserver("browser.content.full-zoom", this);
        } catch(e) {}

        // unobserve pref changes
        this.pref.QueryInterface(Components.interfaces.nsIPrefBranch2).removeObserver("", this);
    },

    log: function(aStr) {
        aStr = "imgfetcher: " + aStr;
        dump(aStr + '\n');
        var consoleServ = Components.classes["@mozilla.org/consoleservice;1"].
             getService(Components.interfaces.nsIConsoleService);
        consoleServ.logStringMessage(aStr);
    },

    /* mouse listeners */
    mouseout: function(evt) {
        imgFetcher.cancelOverTimeout();
    },
    mousemove: function(evt) {
        imgFetcher.maySetTimeout(evt.target);
    },
    mouseover: function(evt) {
        imgFetcher.maySetTimeout(evt.target);
    },

    maySetTimeout: function(target) {
        // returns if we are not over an image
        var nsIImageLoadingContent = Components.interfaces.nsIImageLoadingContent;
        var imgIRequest = Components.interfaces.imgIRequest;

        if ((target instanceof Ci.nsIDOMHTMLAreaElement) &&
            (target.parentNode instanceof Ci.nsIDOMHTMLMapElement) &&
            (target.parentNode.name)) {
            var mapName = target.parentNode.name;
            var imgs = target.ownerDocument.getElementsByTagName('img');
            for (var i = imgs.length; i--> 0; ) {
                var img = imgs[i];
                if (img.useMap && (img.useMap.split('#')[1] == mapName)) {
                    target = img;
                    break;
                }
            }
        }

        if (!(target instanceof nsIImageLoadingContent) || !(target.currentURI) ||
             (target.getRequest(nsIImageLoadingContent.CURRENT_REQUEST) == null) ||
            !(target.getRequest(nsIImageLoadingContent.CURRENT_REQUEST).imageStatus & imgIRequest.STATUS_SIZE_AVAILABLE)) {
            return;
         }

         // return if image is not displayed
         if (target.naturalHeight == 0 && target.naturalWidth == 0) {
             return;
         }

        // returns if we are not inside main browser.
        // we do not display popup over browser chrome.
        var win = target.ownerDocument.defaultView.top;
        var contentWin = getBrowser().contentWindow;
        var windowIsInsideBrowser = false;
        do {
            if (win == contentWin) {
                windowIsInsideBrowser = true;
            }
            win = win.parent;
        } while (win && win != win.parent)
        if (!windowIsInsideBrowser) {
            return;
        }

        this.cancelOverTimeout();
        var self = this;
        this.overtimeout = setTimeout(function() {
            self.openPopup()
        }, this.overdelay);
        this.overImg = target;
    },

    openPopup: function() {
        if (this.isPopupOpen) {
            return;
        }
        var img = this.overImg;
        var doc = img.ownerDocument;

        this.currentImg = this.overImg;
        this.overImg = null;
        this.isPopupOpen = true;

        var popup = doc.createElement("span");
        popup.textContent = this.popupText;
        popup.style.position = "absolute";
        popup.style.padding = "1px 4px 1px 4px";
        popup.style.MozAppearance = "tooltip";
        popup.style.color = "InfoText";
        popup.style.zIndex = "10000";
        popup.style.cursor = "pointer";
        this.popup = popup;

        // we don't known image size before displaying it. First, display it,
        // then modify left and top properties if needed.
        this.showPopup();

        popup.addEventListener("mouseover", this.mouseImgOver, false);
        img.addEventListener("mouseout", this.mouseImgOut, false);
        doc.addEventListener("DOMAttrModified", this.attrmodified, true);

    },

    showPopup: function() {
        if (this.popup.parentNode)
            this.popup.parentNode.removeChild(this.popup);

        var doc = this.popup.ownerDocument;
        if (!doc.body) { //XXX: can this happen ?
            this.log("document has no body (at " + doc.location.href + ')');
            this.closePopup();
            return;
        }

        // we don't known popup dimensions, so we must first display it, then
        // compute it's position according it's size, then place it.
        doc.body.appendChild(this.popup);

        // get popup dimensions
        var popupStyle = doc.defaultView.getComputedStyle(this.popup, "");
        var popupWidth = parseInt(popupStyle.width) + parseInt(popupStyle.paddingLeft) + parseInt(popupStyle.paddingTop);
        var popupHeight = parseInt(popupStyle.height) + parseInt(popupStyle.paddingBottom) + parseInt(popupStyle.paddingTop);

        // checks if image is big enough: if popup is too large compared to
        // image, we close it
        // image dimensions must be superior to popup dimensions
        if ((this.currentImg.height <= popupHeight) || (this.currentImg.width <= popupWidth)) {
            this.closePopup();
            return;
        }
        // at least one image dimension must be two times popup dimensions
        if (!(this.currentImg.height >= popupHeight * 2) && !(this.currentImg.width >= popupWidth * 2)) {
            this.closePopup();
            return;
        }

        var offsetLeft = 0;
        var offsetTop = 0;
        var el = this.currentImg;
        while (el) {
            offsetLeft += el.offsetLeft;
            offsetTop += el.offsetTop;
            el = el.offsetParent;
        }
        var imgStyle = this.currentImg.ownerDocument.defaultView.getComputedStyle(this.currentImg, "");
        offsetLeft += parseInt(imgStyle.getPropertyValue("border-left-width"));
        offsetTop += parseInt(imgStyle.getPropertyValue("border-top-width"));

        switch (this.popupPos) {
            case 1: // upper left
                this.popup.style.left = offsetLeft + 'px';
                this.popup.style.top = offsetTop + 'px';
                break;
            case 2: // upper right
                this.popup.style.left = (offsetLeft + this.currentImg.width - popupWidth) + 'px';
                this.popup.style.top = offsetTop + 'px';
                break;
            case 3: // lower right
                this.popup.style.left = offsetLeft + this.currentImg.width - popupWidth + 'px';
                this.popup.style.top = offsetTop + this.currentImg.height - popupHeight + 'px';
                break;
            case 4: // lower left
                this.popup.style.left = offsetLeft + 'px';
                this.popup.style.top = offsetTop + this.currentImg.height - popupHeight + 'px';
                break;
        }

    },

    saveCurrent: function(evt) {
        if (!imgFetcher.isPopupOpen)
            return;
        if (evt.originalTarget != imgFetcher.popup)
            return;

        var img = imgFetcher.currentImg;
        imgFetcher.closePopup();
        var href = img.currentURI.spec;
        var docUrl = img.ownerDocument.location.href;

        var principal = img.ownerDocument.nodePrincipal;
        urlSecurityCheck(href, principal);
        saveImageURL( href, null, "SaveImageTitle", false, false, makeURI(docUrl));
    },

    cancelOverTimeout: function() {
        if (this.overtimeout) {
            window.clearTimeout(this.overtimeout);
            this.overtimeout = null;
        }
    },

    mouseImgOver: function(evt) {
        imgFetcher.clearCloseTimeout();
        evt.target.removeEventListener("mouseover", imgFetcher.mouseImgOver, false);
        evt.target.addEventListener("mouseout", imgFetcher.mouseImgOut, false);
    },

    mouseImgOut: function(evt) {
        imgFetcher.clearCloseTimeout();
        imgFetcher.closetimeout = setTimeout(function() {imgFetcher.closePopup()}, imgFetcher.closedelay);
        evt.target.removeEventListener("mouseout", imgFetcher.mouseImgOut, false);
        evt.target.addEventListener("mouseover", imgFetcher.mouseImgOver, false);
    },

    attrmodified: function(evt) {
        // if style attribute of image or an ascendant of image has changed,
        // popup position has possibly changed; redraw it
        if (evt.attrName == "style") {
            var target = evt.target;
            var item = imgFetcher.currentImg;
            while (item) {
                if (item == target) {
                    imgFetcher.showPopup();
                    return;
                 }
                 item = item.parentNode;
           }
        }
    },

    clearCloseTimeout: function(evt) {
        if (this.closetimeout) {
            window.clearTimeout(this.closetimeout);
            this.closetimeout = null;
        }
    },

    closePopup: function() {
        if (!this.isPopupOpen) {
            return;
        }
        this.clearCloseTimeout();

        this.currentImg.removeEventListener('mouseout', this.mouseImgOut, false);
        this.popup.removeEventListener("mouseout", this.mouseImgOut, false);
        this.currentImg.removeEventListener('mouseover', this.mouseImgOver, false);
        this.popup.removeEventListener("mouseover", this.mouseImgOver, false);
        this.currentImg.ownerDocument.removeEventListener("DOMAttrModified", this.attrmodified, true);

        this.currentImg = null;
        this.isPopupOpen = false;
        this.hidePopup();
    },

    hidePopup: function() {
        this.popup.parentNode.removeChild(this.popup);
        this.popup = null;
    },

    /* nsIObserver */
    observe: function(subject, topic, data) {
        if (topic != "nsPref:changed" ) {
            return;
        }
        switch(data) {
            case "overdelay":
                this.overdelay = this.pref.getIntPref("overdelay");
            break;
            case "closedelay":
                this.closedelay = this.pref.getIntPref("closedelay");
            break;
            case "popuppos":
                var popupPos = this.pref.getIntPref("popuppos");
                if ([1,2,3,4].some(function(x) { return (popupPos == x) }))
                    this.popupPos = popupPos;
            break;
        }
    },

    /*  nsIWebProgressListener */
    onLocationChange: function(progress, request, uri) {
        // we're leaving a page, close popup
        this.closePopup();
    },
    onProgressChange: function(webProgress , request , curSelfProgress , maxSelfProgress , curTotalProgress , maxTotalProgress) {},
    onStatusChange: function(webProgress , request , stateFlags , status) {},
    onSecurityChange: function(webProgress ,request,state) {},
    onStateChange: function (webProgress, request, state, status) {},


    // if popup is open over an image when zoom value changes, page layout will
    // be modified while popup stays at the same place. So we need to redraw
    // popup after zoom change
    //
    // At the moment, we observe browser.content.full-zoom pref, but it won't
    // work in many cases: (if siteSpecific is false, if zoom.full is false, if
    // an extension sets zoom directly trough markupDocumentViewer. See mozilla
    // #416226 and #303405 
    //
    // nsIContentPrefObserver
    onContentPrefSet: function(group, name, value) {
        this.onZoomPrefChange(group, name);
    },
    onContentPrefRemoved: function(group, name) {
        this.onZoomPrefChange(group, name);
    },
    onZoomPrefChange: function(group, name) {
        if (name != "browser.content.full-zoom")
            return;

        if (!this.isPopupOpen)
            return;

        var grouper = Components.classes["@mozilla.org/content-pref/hostname-grouper;1"].getService(Ci.nsIContentURIGrouper);
        if (grouper.group(getBrowser().currentURI) != group)
            return;

        this.showPopup();
    },

}

function imgFetcherInit() {
    imgFetcher.init()
}

function imgFetcherUninit() {
    imgFetcher.uninit()
}

window.addEventListener("load", imgFetcherInit, false);
window.addEventListener("unload", imgFetcherUninit, false);
